/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File events.c : Event handling code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "z80.h"
#include "events.h"
#include "mem.h"
#include "resource.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "intrface.h"
#include "kbd.h"

#include "mous_int.h"
#include "time_int.h"
#include "audio.h"

#ifdef USE_AIR
#include "airclient2.h"
#endif


#define UCLOCKS_PER_FRAME  ((UCLOCKS_PER_SEC)/(TIMER_SPEED))


//-----------------------------------------------------------------------------


extern volatile int SyncTimer;  //declared in procssor.c

static void SamMouseHandler(int Button, int dx, int dy);

int SAMPointerMode=1;  //pointer attatched and x magnification

static uclock_t uticks;
static uclock_t new_uticks;


//----------------------------------------------------------------------------


/* Check & process any incoming events */
void check_events(void)
{
  static FirstTime=1;

  //updatuje klavesnici a joystick
  KbdUpdate();
  JoyUpdate();

#ifdef USE_AIR
  AIR_UPDATE();
#endif

  /* If this is the first call to check_events(), after */
  /* the first emulation frame, then use the UI to put  */
  /* up a splash banner.	                              */
/*
  if(GetCfg(Scode) && FirstTime) {
    FirstTime = False;
    ui_banner();
    return;
  }
*/
  /* Actually all we do is to call the mouse and keyboard update routines */
  /* which in turn call the appropriate update routines */
  if(SAMPointerMode) sam_mouse_xy_write(0,0);

  MouseUpdate();
  EmulatorControl();
}


void save_screenshot(int border) {
//
//save screenshot to a numbered file
//
  int num;
  char fname[20];

  //najde prvni nepouzite cislo snap%i.pcx
  for(num=0;;num++) {
    FILE *f;
    sprintf(fname,"snap%i.pcx",num);
    f=fopen(fname,"rb");
    if(!f) break;
    fclose(f);
  }

  VideoScreenShot(fname,border);
}


/* Mouse handler - control is passed to here if an event is */
/* pending when MouseUpdate() is called from check_events() */
static void SamMouseHandler(int Button, int dx, int dy)
{
	int SAMButtons;

	static int OldButton = 0xff;

	/* process a pointer move event */
	if (SAMPointerMode && (dx || dy)) {
		dx = (dx*SAMPointerMode);
		dy = - dy;
		sam_mouse_xy_write(dx, dy);
	}

	if (OldButton != Button) {

		OldButton = Button;

		if (Button & MOUSE_LEFTBUTTON)
			SAMButtons&=~SAM_BUTTON_1;
		else
			SAMButtons|= SAM_BUTTON_1;
	
		if (Button & MOUSE_MIDDLEBUTTON)
			SAMButtons&=~SAM_BUTTON_2;
		else
			SAMButtons|= SAM_BUTTON_2;
	
		if (Button & MOUSE_RIGHTBUTTON)
			SAMButtons&=~SAM_BUTTON_3;
		else
			SAMButtons|= SAM_BUTTON_3;

		sam_mouse_button_write(SAMButtons);
	}       
}


//-----------------------------------------------------------------------


void wait_for_timer(void)
{
  /* Wait for the Sync timer to time out SyncTimer is raised */
  /* by the timer signal in emulator_sync_timer()            */
#ifndef __DJGPP
  while(!SyncTimer) sigpause(0);
  
  /* If the sync time is not selected this will set the */
  /* timer flag to True, so there will never be a pause */
  SyncTimer = !GetCfg(sync);
#else
  
  //DJGPP
#ifdef INT_TIMER
  while(!SyncTimer) {
    //    asm __volatile__ ("hlt");
//    AudioUpdate();
    pause();
  }
  SyncTimer=!GetCfg(sync);
  AudioUpdate();
#else
//  AudioUpdate();
  while(!SyncTimer) {
    new_uticks = uclock();
    SyncTimer = ((new_uticks - uticks) >= UCLOCKS_PER_FRAME);
  }
  uticks = new_uticks;
  SyncTimer = !GetCfg(sync);
  AudioUpdate();
#endif
  
#endif
}


//-----------------------------------------------------------------------


void reinit_sync_timer(void) {
//
//(re)initialize sync timer (depends on actual state of GetCfg(sync))
//
//by Aley 2001
//
  static int first_call=1;
  if(first_call) {
    OnQuit(TimerQuit,"TimerQuit()");
    first_call=0;
  }

	/* Initialize the sync timer if required */
	if(cfg.sync) {
#ifdef INT_TIMER
		/* Use a timer interrupt to synchronize the emulator */
		TimerInstall(TIMER_SPEED,(int*)&SyncTimer);
#else
		/* Use uclock to synchronize the emulator */
    extern uclock_t uticks;
    uticks=uclock();
#endif
	} else {
    TimerQuit();
  }
}


//-----------------------------------------------------------------------


void EventsInit(void) {
//
//initialize events.c stuff
//
  int i;

  //mouse initialization
  if(MouseInit(SamMouseHandler)) {
    simcoupe_mesg(SIMCOUPE_WARN,"Can't initialize mouse");
  }
  sam_mouse_init();
  OnQuit(MouseClose,"MouseClose()");

  /* printer is initialized when it is first time used */
}

