/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File intrface.c - User interface implementation in VGA text mode
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <dpmi.h>
#include <string.h>
#include <unistd.h>
#include <vector>
#include <algorithm>
#include <string.h>
#include <unistd.h>

extern "C" {
#include "z80.h"
#include "intrface.h"
#include "version.h"
#include "util.h"
#include "resource.h"
#include "speaker.h"
#include "audio.h"
#include "fdi.h"
#include "sna.h"
#include "video.h"
#include "joy.h"
#include "mem.h"
}


#include "support/FindFile.h"


//-----------------------------------------------------------------------------


#define credits_line "ASCD " VERSION_TEXT " - by (c) Aley Keprt 1998-2001, Based on SimCoupe by A.J.Skillman"

#define MAX_PATH 260

#define SCREEN_H 25

#define MENU_LINES 15

#define gotoat(y,x) gotoxy((x)+1,(y)+1)


char workdir[1000];   //working dir pri spusteni emulatoru


const char *onoff[]={"off","on "};

const char *joylist[]={"off   ","arrows","joy 1 ","joy 2 ","joy 3 ","joy 4 ",
  "joy 5 ","joy 6 ","joy 7 ","joy 8 ","joy 9 ","joy 10"};

const char *memlist[]={"0MB","1MB","2MB","3MB","4MB"};

const char *kbdlist[]={"US","UK"};

const char *modelist[]={"Sam ","48k ","128k"};


enum {
  //ascii kody
  VK_Esc    = 27,
  VK_Backspace=8,
  VK_Enter  = 13,
  VK_Space  = ' ',
  //scan kody
  VK_Pause  = 256+70,
  VK_Up     = 256+72,
  VK_Down   = 256+80,
  VK_Exit   = 256+138,
};


enum MenuReturn {
  MR_Continue,
  MR_Redraw,
  MR_Exit,
  MR_Reinit,
};


struct MenuItem {
  char *text;
  const char *help;
};


MenuItem main_menu[]={
  {"Main Menu"},
  {"File","Up/Down to move cursor, Enter/Space to select, Esc/Backspace to close menu"},
  {"Mode: 12345","Select emulation mode - causes reset of emulated machine!"},
  {"Options","Up/Down to move cursor, Enter/Space to select, Esc/Backspace to close menu"},
  {"Continue","Return to emulator"},
  {"Exit","Exit to DOS"},
  {0,0},
};


enum {
  M_File,
  M_Mode,
  M_Options,
  M_Continue,
  M_Exit,
};


MenuItem file_menu[]={
  {"File"},
  {"Disk 1","Insert Sam disk image into drive 1 (.dsk, .sad, .sad.gz)"},
  {"Disk 2","Insert Sam disk image into drive 2 (.dsk, .sad, .sad.gz)"},
  {"Load Snap/Tape","Load snapshot or tape file (.z80, .sna, .fux, .tap) - supports .gz files"},
//  {"Save Snap","Save ZX Spectrum snapshot (.sna)"},
  {0,0},
};


enum {
  M_Disk1,
  M_Disk2,
  M_LoadSnap,
  M_SaveSnap,
};


MenuItem options_menu[]={
  {"Options"},

  {"Graphics...","Configure graphics [submenu]"},
  {"Joysticks...","Configure joysticks [submenu]"},

  {"Sync: 123","50Hz Synchronization (default: on)"},
  {"External Memory: 0MB","Sam Coupe external memory - needs restart! (default: 0MB)"},
//  {"ROM images","Configure ROM images"},

  {"Keyboard: 12","Switch between US and UK keyboard mode (default: US)"},

  {"Audio: 123","Beeper & SAA1099 & AY8910 Audio emulation (default: on)"},
  {"Speaker: 123","Use PC speaker for ZX beeper emulation (default: off)"},
  {0,0},
};


enum {
  M_Graph,
  M_Joy,

  M_Sync,
  M_ExtMem,
//  M_ROM,

  M_Keyboard,

  M_Audio,
  M_Speaker,

};


MenuItem graph_menu[]={
  {"Graphics"},
  {"VESA: 123","Utilize VESA VBE for hires graphics - needs restart! (default: on)"},
  {"AutoScale: 123","Automatically select best resolution (default: on)"},
  {"Scanlines: 123","Scanlines (faster) in mode 640x480 (default: off)"},
  {"Flash: 123","Attribute flashing emulation (default: on)"},
  {0,0},
};


enum {
  M_Vesa,
  M_Autoscale,
  M_Scanlines,
  M_Flash,
};


MenuItem joy_menu[]={
  {"Joystick Emulation"},
  {"Kempston: 123456","Kempston joystick emulation (default: off)"},
  {"Sinclair 1: 123456","Sinclair joystick 1 (right) emulation (default: off)"},
  {"Sinclair 2: 123456","Sinclair joystick 2 (left) emulation (default: off)"},
  {"Cursor: 123456","Cursor joystick emulation (default: arrows)"},
  {"Fuller: 123456","Fuller joystick emulation (default: off)"},
  {0,0},
};


enum {
  M_Kempston,
  M_Sinclair1,
  M_Sinclair2,
  M_Cursor,
  M_Fuller,
};


static const int back_color=1;                 //barva pozadi
static const int text_color=back_color*16+15;  //barva textu
static const int title_color=back_color*16+14; //barva nadpisu
static const int sel_color=8*16+text_color;    //barva selekce
static const int help_color=7;                 //barva helpu

static int menusel;  //posledni volba po ukonceni menu


extern int flash_phase;


//-----------------------------------------------------------------------------


struct File {
  char *name;
  int dir;

  File() {name=0;dir=0;}
//  ~File() {free((void*)name);name=0;dir=0;}

  bool operator<(const File &f2)const {
    return dir>f2.dir || (dir==f2.dir && stricmp(name,f2.name)<0);
  }
};


std::vector<File> files;


//-----------------------------------------------------------------------------


int getkey() {
//
//cte klavesu, rozsirenym kodum pricita 256
//
  int k=getch();
  if(!k) k=256+getch();
  return k;
}


//-----------------------------------------------------------------------------


void ui_enter(void) {
//
//halt the processor at the next frame interrupt
//
  InterruptType = Z80_menu;
}


//-----------------------------------------------------------------------------


static void setvmode(int mode) {
//
//zapne textovy mod
//
  REGS reg;
  reg.x.ax=mode;
  int86(0x10,&reg,&reg);
}


//-----------------------------------------------------------------------------


void menu(const char *title,int width,int lines,bool help,MenuItem item[],MenuReturn(*handler)(int)) {
//
//obecne menu
//
  int sel=0;
  char *old_screen=0;

  textattr(text_color);

  if(!help) {
    //vypise aktualni adresar
    char cwd[1000];
    getcwd(cwd,1000);
    cwd[67]=0;
    int len=strlen(cwd);
    gotoat(SCREEN_H-3,0);
    textattr(help_color);
    cprintf("%*s%s%*s",(80-len)/2,"",cwd,(80-len)/2,"");
    textattr(text_color);
  }

  const int items=lines;  //items=pocet prvku
  if(lines>MENU_LINES) lines=MENU_LINES; //lines=pocet radku
  int first=0;            //first=prvek na prvnim radku

  const int swidth=width+6;
  const int slines=lines+4;

  //spocita horni levy roh menu
  const int xpos=(80-swidth)/2;
  const int ypos=(SCREEN_H-slines)/2;

  //uchova do bufferu puvodni obsah obrazovky
  if(!old_screen) {
    old_screen=new char[swidth*slines*2];
    gettext(xpos+1,ypos+1,xpos+swidth,ypos+slines,old_screen);
  }

  //pripravi vodorovnou caru
  char buf[81];
  memset(buf,'',swidth);
  buf[0]=buf[swidth-1]=' ';
  buf[swidth]=0;

  //horni okraj
  gotoat(ypos,xpos);
  buf[1]='';
  buf[swidth-2]='';
  cputs(buf);

  //nadpis
  gotoat(ypos+1,xpos);
  cprintf(" ");
  textattr(title_color);
  cprintf(" %-*s ",width,title);
  textattr(text_color);
  cprintf(" ");

  //oddeleni nadpisu
  gotoat(ypos+2,xpos);
  buf[1]='';
  buf[swidth-2]='';
  cputs(buf);

#define print(i) \
  { gotoat(ypos+3+i,xpos);               \
    cprintf(" ");                       \
    if(i==sel) textattr(sel_color);      \
    cprintf(" %-*s ",width,item[first+i].text);\
    if(i==sel) textattr(text_color);     \
    cprintf(" "); }

redraw:
  textattr(text_color);

  //volby menu
  for(int i=0;i<lines;i++) print(i);

  //dolni okraj
  gotoat(ypos+lines+3,xpos);
  buf[1]='';
  buf[swidth-2]='';
  cputs(buf);

  while(1) {

    //oznaci selektnuty radek
    print(sel);

    //vypise help
    if(help) {
      gotoat(SCREEN_H-3,0);
      textattr(help_color);
      int len=strlen(item[sel].help);
      cprintf("%*s%s%*s",(80-len)/2,"",item[sel].help,(80-len)/2,"");
      textattr(text_color);
    }

next_key:
    if(!kbhit()) pause();
    switch(getkey()) {

    case VK_Exit:
      setvmode(3);
      Quit(0);
      break;

    case VK_Esc:
    case VK_Backspace:
      sel=-1;
      goto exit;

    case VK_Enter:
    case VK_Space:
      switch(handler(first+sel)) {
      case MR_Redraw:
        goto redraw;
      case MR_Exit:
        goto exit;
      }
      break;

    case VK_Up:
      if(first>0 && sel==2) {
        //posun okna
        sel++;
        print(sel-1);
        sel--;
        movetext(xpos+4,ypos+4,xpos+3+width,ypos+slines-2,xpos+4,ypos+5);
        first--;
        print(sel-2);
      } else if(sel) {
        sel--;
        print(sel+1);
      }
      break;

    case VK_Down:
      if(first+lines<items && sel==lines-3) {
        //posun okna
        sel++;
        print(sel-1);
        sel--;
        movetext(xpos+4,ypos+5,xpos+3+width,ypos+slines-1,xpos+4,ypos+4);
        first++;
        print(sel+2);
      } else if(sel+1<lines) {
        sel++;
        print(sel-1);
      }
      break;

    default:
      goto next_key;
    }
  }

exit:
  //obnovi puvodni obsah obrazovky
  puttext(xpos+1,ypos+1,xpos+swidth,ypos+slines,old_screen);
  delete old_screen;

  menusel=first+sel;

#undef print
}


//-----------------------------------------------------------------------------


void menu(MenuItem item[],MenuReturn(*handler)(int)) {
//
//klasicke menu
//
  item++;

  //zjisti rozmery menu
  int width=strlen(item[-1].text);
  int lines=0;
  for(;item[lines].text;lines++) {
    int len=strlen(item[lines].text);
    if(len>width) width=len;
  }

  menu(item[-1].text,width,lines,1,item,handler);
}


//-----------------------------------------------------------------------------


MenuReturn dirlist_handler(int sel) {
//
//dirlist handler
//
  return MR_Exit;
}


//-----------------------------------------------------------------------------


void open_file(const char *title, const char *mask, char *result) {

  //uchova si aktualni adresar - obnovi ho, pokud nedojde k otevreni souboru
  char cwd[1000];
  getcwd(cwd,1000);

again:

  for(FindFile find("*.*");find.found();find.next()) {

    if(!strcmp(find.getname(),".")) continue;

    int accept=find.isdir();

    //otestuje konec jmena souboru podle kazde masky
    int len=strlen(find.getname());
    const char *p=mask;
    while(!accept && *p) {
      //najde konec jedne masky
      const char *end;
      for(end=p;*end!=0 && *end!=';';end++);
      int masklen=end-p;

      //porovna konec jmena souboru s maskou
      if(!strncasecmp(find.getname()+len-masklen,p,masklen)) accept=1;

      //posune se na dalsi masku
      p+=masklen+1;
    }

    if(accept) {
      //prida do seznamu
      File file;

      if(find.isdir()) {
        //adresar uzavre do hranatych zavorek
        file.name=(char*)malloc(len+2+1);
        sprintf(file.name,"[%s]",find.getname());
      } else {
        //soubory nemeni
        file.name=strdup(find.getname());
      }
      file.dir=find.isdir();
      files.push_back(file);

      //all uppercase --> lowercase after first letter
      int i;
      for(i=0;file.name[i];i++) {
        if(file.name[i]>='a' && file.name[i]<='z') break;
      }
      if(!file.name[i]) strlwr(file.name+(file.name[0]=='[' ? 2 : 1));
    }
  }

  //seradi
  std::sort(files.begin(),files.end());

  //---------------------------------------------------------------------------

  //udela kopii ve formatu menu
  MenuItem *menutext=new MenuItem[files.size()];
  for(int i=files.size();i--;) menutext[i].text=files[i].name;

  menu(title,44,files.size(),0,menutext,dirlist_handler);

  //uchova si kopii zvolene polozky
  File file;
  if(menusel>=0) {
    file.name=strdup(files[menusel].name);
    file.dir=files[menusel].dir;
  }

  //odalokuje vsechny dynamicke objekty
  for(int i=files.size();i--;) free(files[i].name);
  files.clear();
  delete menutext;

  if(file.dir) {
    //do adresare vleze
    file.name[strlen(file.name)-1]=0;
    chdir(file.name+1);
    free(file.name); //odalokuje dynamickou pamet
    goto again;
  } else if(menusel>=0) {
    //soubor vraci
    strcpy(result,file.name);
    return;
  }

  //obnovi puvodni adresar
  chdir(cwd);
}


//-----------------------------------------------------------------------------


void save_file(const char *title, const char *mask, char *result) {
//
//save-file dialog
//
//entry: title  = dialog's title (caption)
//       mask   = filename extension
//       result = where to store result
//
  strcpy(result,"snap");
  strcat(result,mask);
  return;
}


//-----------------------------------------------------------------------------


MenuReturn file_handler(int sel) {
//
//file_menu handler
//
  char fname[1000]="";

  switch(sel) {
  case M_Disk1:
    open_file("Insert disk 1 (.sad;.dsk;.sad.gz)",".sad;.dsk;.sad.gz",fname);
    if(fname[0]) insert_disk(fname,1);
    break;

  case M_Disk2:
    open_file("Insert disk 2 (.sad;.dsk;.sad.gz)",".sad;.dsk;.sad.gz",fname);
    if(fname[0]) insert_disk(fname,2);
    break;

  case M_LoadSnap:
    open_file("Load Snapshot or Tape (.z80;.sna;.fux;.tap)",
              ".z80;.sna;.slt;.fux;.tap;.z80.gz;.sna.gz;.slt.gz;.fux.gz;.tap.gz",fname);
    if(fname[0]) insert_disk(fname,1);
    break;

  case M_SaveSnap:
    save_file("Save ZXS Snapshot (.sna.gz)",".sna",fname);
    if(fname[0]) save_sna(fname);
    break;
  }

  return MR_Continue;
}


//-----------------------------------------------------------------------------


static inline void init_goption(const int item,int state,const char *list[]=0) {
//
//inicializuje polozku item na hodnotu state
//
  if(!list) {
    list=onoff;
    state=(bool)state;
  }

  char *pos=2+strstr(graph_menu[item+1].text,": ");
  assert(pos);

  strcpy(pos,list[state]);
}


//-----------------------------------------------------------------------------


MenuReturn graph_handler(int sel) {
//
//graph_menu handler
//
  switch(sel) {
/*
  case M_Vesa:
    SetCfg(novesa,!GetCfg(novesa));
    init_goption(M_Vesa,!GetCfg(novesa));
    break;
*/
  case M_Autoscale:
    init_goption(M_Autoscale,cfg.autoscale^=1);
    break;

  case M_Scanlines:
    init_goption(M_Scanlines,cfg.scanlines^=1);
    break;

  case M_Flash:
    init_goption(M_Flash,cfg.flash^=1);
    flash_phase=!cfg.flash; //takhle to bylo i v SimCoupe
    break;

  default:
    return MR_Continue;
  }

  return MR_Redraw;
}


//-----------------------------------------------------------------------------


static inline void init_joy(const int item,int state) {
//
//inicializuje text joysticku item na hodnotu state
//
  char *pos=2+strstr(joy_menu[item+1].text,": ");
  assert(pos);
  strcpy(pos,joylist[state]);
}


//-----------------------------------------------------------------------------


MenuReturn joy_handler(int sel) {
//
//joy_menu handler
//
  switch(sel) {

  case M_Kempston:
    cfg.kempston++;
    if(cfg.kempston>JoyGetNum()) cfg.kempston=-1;
    init_joy(M_Kempston,cfg.kempston+1);
    break;

  case M_Sinclair1:
    cfg.sinclair1++;
    if(cfg.sinclair1>JoyGetNum()) cfg.sinclair1=-1;
    init_joy(M_Sinclair1,cfg.sinclair1+1);
    break;

  case M_Sinclair2:
    cfg.sinclair2++;
    if(cfg.sinclair2>JoyGetNum()) cfg.sinclair2=-1;
    init_joy(M_Sinclair2,cfg.sinclair2+1);
    break;

  case M_Cursor:
    cfg.cursor++;
    if(cfg.cursor>JoyGetNum()) cfg.cursor=-1;
    init_joy(M_Cursor,cfg.cursor+1);
    break;

  case M_Fuller:
    cfg.fuller++;
    if(cfg.fuller>JoyGetNum()) cfg.fuller=-1;
    init_joy(M_Fuller,cfg.fuller+1);
    break;

  default:
    return MR_Continue;
  }

  return MR_Redraw;
}


//-----------------------------------------------------------------------------


static inline void init_option(const int item,int state,const char *list[]=0) {
//
//inicializuje polozku item na hodnotu state
//
  if(!list) {
    list=onoff;
    state=(bool)state;
  }

  char *pos=2+strstr(options_menu[item+1].text,": ");
  assert(pos);

  strcpy(pos,list[state]);
}


//-----------------------------------------------------------------------------


MenuReturn options_handler(int sel) {
//
//options_menu handler
//
  switch(sel) {

  case M_Graph:
    init_goption(M_Vesa,cfg.vesa);
    init_goption(M_Autoscale,cfg.autoscale);
    init_goption(M_Scanlines,cfg.scanlines);
    init_goption(M_Flash,cfg.flash);
    menu(graph_menu,graph_handler);
    break;

  case M_Joy:
    init_joy(M_Kempston,cfg.kempston+1);
    init_joy(M_Sinclair1,cfg.sinclair1+1);
    init_joy(M_Sinclair2,cfg.sinclair2+1);
    init_joy(M_Cursor,cfg.cursor+1);
    init_joy(M_Fuller,cfg.fuller+1);
    menu(joy_menu,joy_handler);
    break;

  case M_Sync:
    init_option(M_Sync,cfg.sync^=1);
    reinit_sync_timer();
    break;

  case M_Keyboard:
    cfg.uk^=1;
    init_option(M_Keyboard,cfg.uk,kbdlist);
    break;

  case M_Audio:
    init_option(M_Audio,cfg.audio^=1);
    if(cfg.audio) AudioInit(); else AudioClose();
    break;

  case M_Speaker:
    init_option(M_Speaker,cfg.use_speaker^=1);
    if(cfg.use_speaker) SpeakerOn(); else SpeakerOff();
    break;

  default:
    return MR_Continue;
  }

  return MR_Redraw;
}


//-----------------------------------------------------------------------------


MenuReturn main_handler(int sel) {
//
//main_menu handler
//
  switch(sel) {

  case M_File:
    menu(file_menu,file_handler);
    break;

  case M_Mode:
    cfg.zxmode++;
    if(cfg.zxmode>=3) cfg.zxmode=(EM)0;
    strcpy(main_menu[M_Mode+1].text+6,modelist[cfg.zxmode]);
    set_emulation_mode((EM)cfg.zxmode);
    break;

  case M_Options:
    init_option(M_Sync,cfg.sync);
    init_option(M_ExtMem,cfg.extmem,memlist);

    init_option(M_Keyboard,cfg.uk,kbdlist);

    init_option(M_Audio,cfg.audio);
    init_option(M_Speaker,cfg.use_speaker);

    menu(options_menu,options_handler);
    break;

  case M_Continue:
    return MR_Exit;

  case M_Exit:
    setvmode(3);
    Quit(0);
    break;
  }

  return MR_Continue;
}


//-----------------------------------------------------------------------------


static void draw_help(void) {
//
//vypise (de facto na pozadi) help
//
  gotoat(0,0);

  textattr(6);
  cprintf("Function keys:\r\n");
  textattr(8);
  cprintf(
//    "F1=Help?\r\n"
    "F2=Main Menu\r\n"
//    "F3=.\r\n"
//    "F4=.\r\n"
    "F5=Swich Resolution 320x240/640x480\r\n"
    "Ctrl+F5=Autoscale on/off (automatically selects best resolution)\r\n"
    "F6=Mouse on/off\r\n"
    "Ctrl+F6=Mouse Resolution 1x/2x\r\n"
    "F7=Keyboard Mode Sam/ZXS\r\n"
//    "F8=.\r\n"
    "F9=Synchronization on/off\r\n"
    "F10=Save PCX Snapshot\r\n"
    "Ctrl+F10=Save PCX Snapshot With Border\r\n"
    "F11=NMI\r\n"
    "F12=Reset\r\n"
    "Ctrl+F12=Exit Emulator\r\n"
    "\r\n\n\n"
    "Type \"ascd -help |more\" to see the complete list of command line switches.\r\n"
    "Type \"ascd -help >list\" to put the complete list of switches to a file.\r\n"
  );
}


//-----------------------------------------------------------------------------


void ui_run(void) {
//
//vykona menu
//
//volano pouze z procssor.c pri preruseni
//
  //uchova aktualni adresar
  char cwd[1000];
  getcwd(cwd,sizeof(cwd));

  //pri prvnim spusteni workdir=cwd, potom obnovuje posledni adresar
  if(workdir[0]==0) strcpy(workdir,cwd);
  chdir(workdir);

  VideoTextMode();

  _setcursortype(_NOCURSOR);
  directvideo=1;
  _wscroll=0;

  //vypne blikn textmdu
  union REGS reg;
  reg.x.ax=0x1003;
  reg.x.bx=0;
  int86(0x10,&reg,&reg);

  gotoat(SCREEN_H-1,(80-strlen(credits_line))/2);
  textattr(6);
  cprintf(credits_line);

  draw_help();

  strcpy(main_menu[M_Mode+1].text+6,modelist[cfg.zxmode]);
  menu(main_menu,main_handler);

  getcwd(workdir,sizeof(workdir));
  chdir(cwd);

  //empty the keyboard buffer and return
  while(kbhit()) getch();
}

